local metadata = {
"## Interface: FS15 1.3.0.1 1.3RC1",
"## Title: FabrikScript",
"## Notes: FabrikScript zum erstellen eines Fabrikgebäudes mit verschiedenen Rohstoffen und Produkten",
"## Author: Marhu",
"## Version: 1.4.3-459",
"## Date: 05.08.2015",
"## Web: http://marhu.net"
}
 
local DebugEbene = 0;
local function Debug(printDebug,...)
	if printDebug <= DebugEbene then
		local text, TITLE, VERSION = "","",""
		for i = 1, select("#", ...) do
			if type(select(i, ...)) == "boolean" then
				text = text..(select(i, ...) and "true" or "false").." ";
			else text = text..tostring(select(i, ...) or "nil").." "; end;
		end
		for i = 1, table.getn(metadata) do
			local nBeginn, nEnde = string.find(metadata[i],"## Title: ");
			if nEnde then TITLE = string.sub (metadata[i], nEnde); end;
			local nBeginn, nEnde = string.find(metadata[i],"## Version: ");
			if nEnde then VERSION = " v"..string.sub (metadata[i], nEnde); end;
		end;
		print(TITLE..VERSION..": "..text);
	end;
end;

FabrikScript = {};
FabrikScript.ModDir = g_currentModDirectory
FabrikScript_mt = nil

local nBeginn, nEnde = string.find(FabrikScript.ModDir,"placeable");
if nBeginn then
	FabrikScript_mt = Class(FabrikScript, Placeable);
else
	FabrikScript.RunAsGE = true;
	FabrikScript_mt = Class(FabrikScript, Object);
end

InitObjectClass(FabrikScript, "FabrikScript");

function FabrikScript.onCreate(id)
	local object = FabrikScript:new(g_server ~= nil, g_client ~= nil)
	g_currentMission:addOnCreateLoadedObject(object);
	if object:load(id) then
		g_currentMission:addOnCreateLoadedObjectToSave(object);
        object:register(true);
		Debug(1,string.format("FabrikScript.onCreate(%d) load %s",id,getName(id)));
    else
        object:delete();
    end;
		
end;

function FabrikScript:new(isServer, isClient, customMt)
  
	local mt = customMt;
    if mt == nil then
          mt = FabrikScript_mt;
    end;
  
	local self = {};
	if FabrikScript.RunAsGE then
		self = Object:new(isServer, isClient, mt)
	else
		self = Placeable:new(isServer, isClient, mt);
		registerObjectClassName(self, "FabrikScript");
	end;
	
	self.updateIntervall = 5;	
	self.updateMs = 60000;
	self.updateMin = 0 + self.updateIntervall;
	

			
	return self;
end;
 
function FabrikScript:load(xmlFilename, x,y,z, rx,ry,rz, initRandom)

	self.FehlerText = {}
	self.StoffeIdToName = {}	
	self.trailerTipTrigger = {}
	if not self.RunAsGE then
		if not FabrikScript:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, initRandom) then
			return false;
		end;
		return true;
	else
		self.nodeId = xmlFilename;
		if not self:finalizePlacement() then return false; end;
	end;
	
	return true;
end;

function FabrikScript:finalizePlacement(x, y, z, rx, ry, rz, initRandom)
	
	if not self.RunAsGE then
		FabrikScript:superClass().finalizePlacement(self)
	end
	
	self.ProduktPerHour = Utils.getNoNil(getUserAttribute(self.nodeId, "ProduktPerHour"),1000);
	local StoffeId = 1;	
	local WasserTriggerId = 0;
	local SiloTriggerId = 0;
	
	local InputIndex = getUserAttribute(self.nodeId, "inputIndex");
	if InputIndex then
		local InputId = Utils.indexToObject(self.nodeId, InputIndex);	
		if InputId then
			self.InputName = getName(InputId)
			local numChildren = getNumOfChildren(InputId);
			if 0 >= numChildren then
				Debug(-1,string.format("ERROR: no Resources in InputId %d Name %s",InputId,self.InputName));
				return false;
			else
				Debug(2,string.format("InputId %d Name %s",InputId,self.InputName));
			end
			
			self.Rohstoffe = {}
			
			for i = 1, numChildren do
				local rohstoffId = getChildAt(InputId,i-1)
				local name = Utils.getNoNil(getUserAttribute(rohstoffId, "name"),getName(rohstoffId));
				if self.Rohstoffe[name] == nil then
					self.Rohstoffe[name]={}
					self.Rohstoffe[name].id = StoffeId;
					self.Rohstoffe[name].name = name
					self.Rohstoffe[name].nameL = g_i18n:hasText(name) and g_i18n:getText(name) or name;
					self.StoffeIdToName[StoffeId] = name;
					self.FehlerText[StoffeId] = self.Rohstoffe[name].nameL.." "..(g_i18n:hasText("leer") and g_i18n:getText("leer") or "leer");
					StoffeId = StoffeId + 1;
					self.Rohstoffe[name].fillLevel = 0;
					self.Rohstoffe[name].capacity = Utils.getNoNil(getUserAttribute(rohstoffId, "capacity"),10000);
					self.Rohstoffe[name].factor = Utils.getNoNil(getUserAttribute(rohstoffId, "factor"),1);
					self.Rohstoffe[name].acceptedFillTypes = {}
					local fillTypes = getUserAttribute(rohstoffId, "fillTypes");
					if fillTypes == nil then
						fillTypes = getUserAttribute(rohstoffId, "fruitTypes");
					end
					if fillTypes ~= nil then
						local types = Utils.splitString(" ", fillTypes);
						for k,v in pairs(types) do
							local desc = Fillable.fillTypeNameToDesc[v];
							if desc ~= nil then
								self.Rohstoffe[name].acceptedFillTypes[desc.index] = true;
								Debug(3,string.format("Input %s accept FillType %s",name,desc.name));
							else
								Debug(-1,string.format("ERROR: invalid fillType %s in %s",tostring( v or "nil"),name));
							end;
						end;
					end;
					
					local numTrigger = 0;
					
					local TipTriggerId = getChild(rohstoffId,"TipTrigger")
					if TipTriggerId and TipTriggerId ~= 0 then
								
						local rohstoffTipTrigger = FillableTipTrigger:new(self.isServer, self.isClient)

						if rohstoffTipTrigger:load(TipTriggerId, self.Rohstoffe[name]) then
							self.Rohstoffe[name].TipTrigger = rohstoffTipTrigger
							self.Rohstoffe[name].TipTrigger:register(true)
							self.Rohstoffe[name].allowFillType = function(...) return self:allowFillType(...) end;
							self.Rohstoffe[name].getCapacity = function(...) return self:getCapacity(...) end;
							self.Rohstoffe[name].getFillLevel = function(...) return self:getFillLevel(...) end;
							self.Rohstoffe[name].setFillLevel = function(...) return self:setFillLevel(...) end;
							self.Rohstoffe[name].resetFillLevelIfNeeded = function(...) return self:resetFillLevelIfNeeded(...) end;
							self.Rohstoffe[name].TipTriggerCallback = function(...) return self:TipTriggerCallback(...) end;
							self.Rohstoffe[name].notMoveBack = getUserAttribute(TipTriggerId, "notMoveBack")
							if rohstoffTipTrigger.movingId ~= nil then
								local org_updateMoving = rohstoffTipTrigger.updateMoving
								rohstoffTipTrigger.updateMoving = function (self, delta)
									if not self.isServer and self.isClient then
										local x, y, z = getTranslation(self.movingId)
										local newY = math.min(y + delta*self.moveScale, self.moveMaxY)
										setTranslation(self.movingId, x, newY, z);
									end
									return org_updateMoving(self, delta)
								end
								if self.Rohstoffe[name].notMoveBack == true then
									rohstoffTipTrigger.moveBackScale = 0;
									rohstoffTipTrigger.moveScale = (rohstoffTipTrigger.moveMaxY-rohstoffTipTrigger.moveMinY) / self.Rohstoffe[name].capacity;
								end
							end
							
							removeTrigger(self.Rohstoffe[name].TipTrigger.triggerId);
							addTrigger(self.Rohstoffe[name].TipTrigger.triggerId , "TipTriggerCallback", self.Rohstoffe[name]);
							
							Debug(3,string.format("Input %s load TipTrigger %d",name,TipTriggerId));
							numTrigger = numTrigger + 1;
						else
							rohstoffTipTrigger:delete()
						end

					end
					
					local BaleTriggerId = getChild(rohstoffId,"BaleTrigger")
					if BaleTriggerId and BaleTriggerId ~= 0 then
						self.Rohstoffe[name].BaleTrigger = BaleTriggerId
						addTrigger(self.Rohstoffe[name].BaleTrigger, "BaleTriggerCallback", self);
						Debug(3,string.format("Input %s add BaleTrigger %d",name,BaleTriggerId));
						numTrigger = numTrigger + 1;
					end
					
					local ShovelTargetId = getChild(rohstoffId,"ShovelTarget")
					if ShovelTargetId and ShovelTargetId ~= 0 and self.isServer then
						self.Rohstoffe[name].ShovelTarget = ShovelTarget:new();
						self.Rohstoffe[name].ShovelTarget.Rohstoff = name;
						self.Rohstoffe[name].ShovelTarget.nodeId = ShovelTargetId;
						self.Rohstoffe[name].ShovelTarget.fillTypes = self.Rohstoffe[name].acceptedFillTypes
						g_currentMission:addNodeObject(self.Rohstoffe[name].ShovelTarget.nodeId, self.Rohstoffe[name].ShovelTarget);
						g_currentMission:addNonUpdateable(self.Rohstoffe[name].ShovelTarget);
						self.Rohstoffe[name].ShovelTarget.addShovelFillLevel = function(target, shovel, fillLevelDelta, fillType) return self:addShovelFillLevel(target, shovel, fillLevelDelta, fillType); end;
						Debug(3,string.format("Input %s create ShovelTarget %d",name,ShovelTargetId));
						numTrigger = numTrigger + 1;
					end
					
					local PalletTriggerId = getChild(rohstoffId,"PalletTrigger")
					if PalletTriggerId and PalletTriggerId ~= 0 and self.isServer then
						self.Rohstoffe[name].PalletTrigger = PalletTriggerId
						addTrigger(self.Rohstoffe[name].PalletTrigger, "PalletTriggerCallback", self);
						Debug(3,string.format("Input %s add PalletTrigger %d",name,PalletTriggerId));
						numTrigger = numTrigger + 1;
					end
					
					local WoodTriggerId = getChild(rohstoffId,"WoodTrigger")
					if WoodTriggerId and WoodTriggerId ~= 0 then
						self.Rohstoffe[name].WoodTrigger = WoodTriggerId
						addTrigger(self.Rohstoffe[name].WoodTrigger, "WoodTriggerCallback", self);
						local numLogs = getNumOfChildren(WoodTriggerId);
						if numLogs >= 1 then -- v1.3.3 fix (TracMax)
							self.Rohstoffe[name].logs = {}
							for i = 1, numLogs do 
								local WoodLog = getChildAt(WoodTriggerId,i-1)
								self.Rohstoffe[name].logs[i]={node = WoodLog,rigidBody = getRigidBodyType(WoodLog)}
								setRigidBodyType(WoodLog,"NoRigidBody")
								setVisibility(WoodLog,false)
							end
						end
						Debug(3,string.format("Input %s add WoodTrigger %d",name,WoodTriggerId));
						numTrigger = numTrigger + 1;
					end
					
					local WaterTriggerId = getChild(rohstoffId,"WaterTrigger")
					if WaterTriggerId and WaterTriggerId ~= 0 then
						if WaterFillTrigger then
							local WaterTrigger = getChildAt(WaterTriggerId,0)
							local WaterTriggerFunktionen = {}
							WaterTriggerFunktionen.getCapacity = function(...) return self:getCapacity(self.Rohstoffe[name],...) end;
							WaterTriggerFunktionen.getFillLevel = function(...) return self:getFillLevel(self.Rohstoffe[name],...) end;
							WaterTriggerFunktionen.setFillLevel = function(...) return self:setFillLevel(self.Rohstoffe[name],...) end;
							WaterTriggerFunktionen.Parent = self;
							WasserTriggerId = WasserTriggerId + 1;
							local Trigger = WaterFillTrigger:new(g_server ~= nil, g_client ~= nil)
							Trigger:load(WaterTrigger,WaterTriggerFunktionen)
							Trigger.WasserTriggerId = WasserTriggerId;
							local movingIndex = getUserAttribute(WaterTriggerId, "movingIndex")
							if movingIndex then
								local minY, maxY = Utils.getVectorFromString(getUserAttribute(WaterTriggerId, "moveMinMaxY"));
								if minY ~= nil and maxY ~= nil then
									local maxAmount = tonumber(getUserAttribute(WaterTriggerId, "moveMaxAmount")) or self.Rohstoffe[name].capacity;
									if maxAmount ~= nil then
										Trigger.movingIndex = Utils.indexToObject(WaterTriggerId, movingIndex);
										Trigger.moveMinY = minY;
										Trigger.moveMaxY = maxY;
										Trigger.moveMaxAmount = maxAmount;
									end;
								end;
							end;
							self.Rohstoffe[name].WaterTrigger = Trigger;
							Debug(3,string.format("Input %s add WaterTrigger %d",name,WaterTriggerId));
						else
							Debug(-1,string.format("Input %s WaterTrigger cant find WaterFillTrigger Script",name));
						end
						numTrigger = numTrigger + 1;
					end
					
					if numTrigger == 0 then
						Debug(-1,string.format("ERROR: can not find Trigger in Input %s",name));
						return false;
					end
				else
					Debug(-1,string.format("WARNING: Input %s already exists!",name));
				end
			end
		else
			Debug(-1,string.format("ERROR: no Objekt on InputIndex %s",tostring(InputIndex)));
			return false;
		end
	else
		Debug(-1,"ERROR: InputIndex = nil");
		return false;
	end
	
	local OutputIndex = getUserAttribute(self.nodeId, "outputIndex");
	if OutputIndex then
		local OutputId = Utils.indexToObject(self.nodeId, OutputIndex);	
		if OutputId then
			self.OutputName = getName(OutputId)
			local numChildren = getNumOfChildren(OutputId);
			if 0 >= numChildren then
				Debug(-1,string.format("ERROR: no Produkt in OutputId %d Name %s",OutputId,self.OutputName));
				return false;
			else
				Debug(2,string.format("OutputId %d Name %s",OutputId,self.OutputName));
			end
			
			self.Produkte = {}
			
			for i = 1, numChildren do
				local produktId = getChildAt(OutputId,i-1)
				local name = Utils.getNoNil(getUserAttribute(produktId, "name"),getName(produktId));
				if self.Produkte[name] == nil then
					self.Produkte[name]={}
					self.Produkte[name].id = StoffeId;
					self.Produkte[name].name = name
					self.Produkte[name].nameL = g_i18n:hasText(name) and g_i18n:getText(name) or name;
					self.StoffeIdToName[StoffeId] = name;
					self.FehlerText[StoffeId] = self.Produkte[name].nameL.." "..(g_i18n:hasText("voll") and g_i18n:getText("voll") or "voll");
					StoffeId = StoffeId + 1
					self.Produkte[name].fillLevel = 0;
					self.Produkte[name].capacity = Utils.getNoNil(getUserAttribute(produktId, "capacity"),10000);
					self.Produkte[name].factor = Utils.getNoNil(getUserAttribute(produktId, "factor"),1);
					self.Produkte[name].acceptedFillTypes = {}	
					
					local fillType = getUserAttribute(produktId, "fillType");
					local desc = Fillable.fillTypeNameToDesc[fillType];
					if desc ~= nil then
						
						self.Produkte[name].acceptedFillTypes[desc.index] = true;
						Debug(3,string.format("Output %s accept FillType %s",name,desc.name));
										
						local numTrigger = 0;
						
						local TriggerId = getChild(produktId,"siloTrigger")
						if TriggerId and TriggerId ~= 0 then
							local trigger = SiloTrigger:new(self.isServer, self.isClient);
							
							if not self.RunAsGE and g_currentMission.baseDirectory == "" then
								local org_Filename = getUserAttribute(TriggerId, "particleSystemFilename")
								if org_Filename and Utils.startsWith(org_Filename, "../") then
									setUserAttribute(TriggerId,"particleSystemFilename","String",g_modsDirectory..org_Filename:sub(3))
									org_Filename = getUserAttribute(TriggerId, "particleSystemFilename")
									Debug(3,string.format("set UserAttribute particleSystemFilename to: %s",tostring(org_Filename)));
								end;
							end;
							trigger:load(TriggerId);
							trigger:register(true);
							trigger.fillType = desc.index
							trigger.notfillDone = Utils.getNoNil(getUserAttribute(TriggerId, "notfillDone"),true);
							function trigger:update(dt) end;
							SiloTriggerId = SiloTriggerId + 1;
							trigger.TriggerId = SiloTriggerId;
							self.Produkte[name].SiloTrigger = trigger
							Debug(3,string.format("Output %s add siloTrigger %d",name,TriggerId));
							numTrigger = numTrigger + 1;
						end
						
						local HeapId = getChild(produktId,"Heap")
						if HeapId and HeapId ~= 0 then
							local heap = {}
							heap.id = HeapId;
							local minY, maxY = Utils.getVectorFromString(getUserAttribute(HeapId, "moveMinMaxY"));
							if minY ~= nil and maxY ~= nil then
								local maxAmount = tonumber(getUserAttribute(HeapId, "moveMaxAmount")) or self.Produkte[name].capacity;
								if maxAmount ~= nil then
									heap.moveMinY = minY;
									heap.moveMaxY = maxY;
									heap.moveMaxAmount = maxAmount;
								end;
							end;
							if self.isServer then
								local ShovelTrigger = getChildAt(HeapId,0)
								if ShovelTrigger then
									local trigger = ShovelFillTrigger:new();
									if trigger:load(ShovelTrigger, desc.index) then
										g_currentMission:addUpdateable(trigger);
										heap.ShovelTrigger = trigger
										heap.ShovelTrigger.Produkt = name;
										heap.ShovelTrigger.fillShovel = function(ST, shovel, dt) self:fillShovel(ST, shovel, dt); end;
										Debug(3,string.format("Output %s add ShovelTrigger %d",name,ShovelTrigger));
									else
										Debug(-1,string.format("ERROR: Output %s can not load ShovelTrigger %d",name,ShovelTrigger));
										trigger:delete();
									end;
								end;
							end;
							self.Produkte[name].Heap = heap;
							Debug(3,string.format("Output %s add Heap %d",name,HeapId));
							numTrigger = numTrigger + 1;
						end
						
						local palletSpawnerId = getChild(produktId,"palletSpawner")
						if palletSpawnerId and palletSpawnerId ~= 0 then
							self.FehlerText[StoffeId] = self.Produkte[name].nameL.." "..(g_i18n:hasText("palletPlaceNotFree") and g_i18n:getText("palletPlaceNotFree") or "palletPlaceNotFree");
							StoffeId = StoffeId + 1
							if self.isServer then
								local pallet = {}
								pallet.Filename = getUserAttribute(palletSpawnerId, "palletFilename");
								pallet.Filename = Utils.getFilename(pallet.Filename, self.ModDir)
								pallet.SpawnerTriggerId = getChildAt(palletSpawnerId, 0)

								addTrigger(pallet.SpawnerTriggerId, "palletSpawnerTriggerCallback", self)
								g_currentMission:addNodeObject(pallet.SpawnerTriggerId, self);

								pallet.SpawnerPlaceId = getChildAt(palletSpawnerId, 1)
								pallet.numObjectsInPalletSpawnerTrigger = 0;
								pallet.numObjectsPerID = {};
								self.Produkte[name].palletSpawner = pallet;
								Debug(3,string.format("Output %s add palletSpawner %d",name,palletSpawnerId));
							end;
							numTrigger = numTrigger + 1;
						end;
						if numTrigger == 0 then
							Debug(-1,string.format("ERROR: can not find Trigger in Output %s",name));
							return false;
						end
					else
						Debug(-1,string.format("ERROR: invalid fillType %s in %s",tostring( fillType or "nil"),name));
						return false;
					end;
				else
					Debug(-1,string.format("WARNING: Output %s already exists!",name));
				end;
			end;
		else
			Debug(-1,string.format("ERROR: no Objekt on OutputIndex %s",tostring(OutputIndex)));
			return false;
		end
	else
		Debug(-1,"ERROR: OutputIndex = nil");
		return false;
	end
	
	local PlayerTriggerIndex = getUserAttribute(self.nodeId,"PlayerIndex");
	if PlayerTriggerIndex then
		local PlayerTrigger = Utils.indexToObject(self.nodeId, PlayerTriggerIndex);
		if PlayerTrigger then
			self.PlayerTrigger = PlayerTrigger;
			addTrigger(self.PlayerTrigger, "PlayerTriggerCallback", self);
		end;
	end;
	
	local DoorsIndex = getUserAttribute(self.nodeId, "DoorsIndex");
	if DoorsIndex ~= nil then
		local Doors = Utils.indexToObject(self.nodeId, DoorsIndex);
		if Doors then
			self.Doors = {}
			local numChildren = getNumOfChildren(Doors);
			for i=1,numChildren do
				local Child = getChildAt(Doors, i-1)
				self.Doors[i] = {}
				self.Doors[i].minTrans = Utils.getNoNil(getUserAttribute(Child, "MinTrans"),0);
				self.Doors[i].maxTrans = Utils.getNoNil(getUserAttribute(Child, "MaxTrans"),2);
				self.Doors[i].Trans = self.Doors[i].minTrans;
				local transTime = Utils.getNoNil(getUserAttribute(Child, "TransTime"),1);
				self.Doors[i].transTime = ((self.Doors[i].maxTrans-self.Doors[i].minTrans) / transTime) * 0.001
				local doorGroup = Child;
				local doorIndex = getUserAttribute(Child, "DoorIndex");
				if doorIndex then
					doorGroup = Utils.indexToObject(Child, doorIndex);
				end;
				local numDoors = getNumOfChildren(doorGroup);
				self.Doors[i].door = {}
				for j=1,numDoors do
					self.Doors[i].door[j] = getChildAt(doorGroup, j-1);
				end;
				local triggerIndex = getUserAttribute(Child, "triggerIndex");
				local trigger = Child;
				if triggerIndex then
					trigger = Utils.indexToObject(Child,triggerIndex);
				end;
				self.Doors[i].triggerId = trigger;
				self.Doors[i].entred = 0;
				addTrigger(trigger, "doorTriggerCallback", self);
			end;
		end;
	end;
	
	if self.isClient then
		local WorkAnimationIndex = getUserAttribute(self.nodeId,"WorkAniIndex");
		if WorkAnimationIndex ~= nil and  WorkAnimationIndex ~= "" then
			local WorkAnimation = Utils.indexToObject(self.nodeId, WorkAnimationIndex);
			if WorkAnimation then
				Debug(3,string.format("Find WorkAnimation on Index %s",WorkAnimationIndex));
				self.WorkAnimation = {}
				local PSIndex = getUserAttribute(WorkAnimation,"PartikleIndex");
				if PSIndex ~= nil and PSIndex ~= "" then
					local PSGroup = Utils.indexToObject(WorkAnimation, PSIndex);
					if PSGroup then
						Debug(3,string.format("Find WorkAnimation PS on Index %s",PSIndex));
						self.WorkAnimation.PS = {}
						local numChildren = getNumOfChildren(PSGroup);
						for i = numChildren, 1, -1 do
							local child = getChildAt(PSGroup,i-1)
							local particleSystem = getUserAttribute(child, "particleSystemFilename");
							Debug(3,string.format("WorkAnimation PS %s",tostring(particleSystem)));
							if particleSystem then
								self.WorkAnimation.PS[i] = {}
								local psData = {};
								psData.psFile = particleSystem;
								psData.posX, psData.posY, psData.posZ = getTranslation(child);
								psData.forceNoWorldSpace = true;
								Utils.loadParticleSystemFromData(psData, self.WorkAnimation.PS[i], nil, false, nil, self.ModDir, getParent(child));
								Debug(3,string.format("WorkAnimation PS Emit %s",tostring(self.WorkAnimation.PS[i].isEmitting)));
								local intervall = getUserAttribute(child, "IntervallSek");
								if intervall and intervall > 0 then
									self.WorkAnimation.PS[i].intervall = intervall * 1000;
									self.WorkAnimation.PS[i].elapsed = 0 + self.WorkAnimation.PS[i].intervall;
								end
							end
						end	
					end;
				end
				local AniIndex = getUserAttribute(WorkAnimation,"AnimationIndex");
				if AniIndex ~= nil and AniIndex ~= "" then
					local AniGroup = Utils.indexToObject(WorkAnimation, AniIndex);
					if AniGroup then
						Debug(3,string.format("Find WorkAnimation Ani on Index %s",AniIndex));
						self.WorkAnimation.Ani = {}
						local numChildren = getNumOfChildren(AniGroup);
						for i = numChildren, 1, -1 do
							local child = getChildAt(AniGroup,i-1)
							local ClipName = getUserAttribute(child, "ClipName");
							local MeshIndex = getUserAttribute(child, "MeshIndex");
							local PosIndex = getUserAttribute(child, "PosIndex");
							Debug(3,string.format("WorkAnimation Clip %s",tostring(ClipName)));
							if ClipName then
								self.WorkAnimation.Ani[i] = {}
							
								if MeshIndex and PosIndex then
									local position = Utils.indexToObject(child, PosIndex);
									local pos = {getWorldTranslation(position)}
									local rot = {getWorldRotation(position)}
									
									local child2 = clone(child, true)
									link(getRootNode(),child2); 
									
									local Mesh = Utils.indexToObject(child2, MeshIndex);
									setTranslation(Mesh,unpack(pos))
									setRotation(Mesh,unpack(rot))
									
									self.WorkAnimation.Ani[i].root = child2;
									
									delete(child)
									child = child2;
								end
								
								self.WorkAnimation.Ani[i].Animi = getAnimCharacterSet(child);
								self.WorkAnimation.Ani[i].Clip = getAnimClipIndex(self.WorkAnimation.Ani[i].Animi,ClipName)
								assignAnimTrackClip(self.WorkAnimation.Ani[i].Animi, 0, self.WorkAnimation.Ani[i].Clip);
								setAnimTrackLoopState(self.WorkAnimation.Ani[i].Animi, 0, false);
								setAnimTrackSpeedScale(self.WorkAnimation.Ani[i].Animi, 0, 1);
								self.WorkAnimation.Ani[i].IgnoreDuration = getUserAttribute(child, "IgnoreDuration");
							end;
						end;
					end;
				end;
				local SoundIndex = getUserAttribute(WorkAnimation,"SoundIndex");
				if SoundIndex ~= nil and SoundIndex ~= "" then
					local SoundGroup = Utils.indexToObject(WorkAnimation, SoundIndex);
					if SoundGroup then
						Debug(3,string.format("Find WorkAnimation Sound on Index %s",SoundIndex));
						self.WorkAnimation.Sound = {}
						local numChildren = getNumOfChildren(SoundGroup);
						for i = numChildren, 1, -1 do	
							self.WorkAnimation.Sound[i] = {}
							self.WorkAnimation.Sound[i].node = getChildAt(SoundGroup,i-1)
							setVisibility(self.WorkAnimation.Sound[i].node,false)
							local intervall = getUserAttribute(self.WorkAnimation.Sound[i].node, "IntervallSek");
							if intervall and intervall > 0 then
								self.WorkAnimation.Sound[i].intervall = intervall * 1000;
								self.WorkAnimation.Sound[i].elapsed = 0 + self.WorkAnimation.Sound[i].intervall;
							end
						end
					end
				end	
				local ShaderIndex = getUserAttribute(WorkAnimation,"ShaderIndex");
				if ShaderIndex ~= nil and ShaderIndex ~= "" then
					local ShaderGroup = Utils.indexToObject(WorkAnimation, ShaderIndex);
					if ShaderGroup then
						Debug(3,string.format("Find WorkAnimation Shader on Index %s",ShaderIndex));
						self.WorkAnimation.Shader = {}
						local numChildren = getNumOfChildren(ShaderGroup);
						for i = numChildren, 1, -1 do	
							local ShaderChild = getChildAt(ShaderGroup,i-1)
							local ShaderObjekt = Utils.indexToObject(ShaderChild,getUserAttribute(ShaderChild,"ShaderObjektIndex"));
							if ShaderObjekt == nil then	ShaderObjekt = ShaderChild;	end;
							local SP = {}
							SP.node = ShaderObjekt
							SP.Parameter = getUserAttribute(ShaderObjekt,"parameterName");
							local on,off = {},{};
							on.x,on.y,on.z,on.w = Utils.getVectorFromString(getUserAttribute(ShaderObjekt,"value"));
							off.x,off.y,off.z,off.w = getShaderParameter(SP.node,SP.Parameter);
							local shared = getUserAttribute(ShaderObjekt,"shared");
							if shared == true then
								SP.on = on;	SP.off = off;
							else
								SP.on = off; SP.off = on;
							end
							SP.shared = false;
							setShaderParameter(SP.node,SP.Parameter,SP.off.x,SP.off.y,SP.off.z,SP.off.w,false)	
							self.WorkAnimation.Shader[i] = SP;
						end
					end
				end	
			end
		end
	end;
	
	if not self.RunAsGE then
		local AreaIndex = getUserAttribute(self.nodeId,"Area");
		if AreaIndex ~= nil then         
			local AreaStart = Utils.indexToObject(self.nodeId, AreaIndex);
			if AreaStart ~= nil then
				self.Area = {}
				self.Area.start  = AreaStart;
				self.Area.width	 = getChildAt(AreaStart,0);
				self.Area.height = getChildAt(AreaStart,1);
				local x,_,z = getWorldTranslation(self.Area.start);
				local x1,_,z1 = getWorldTranslation(self.Area.width);
				local x2,_,z2 = getWorldTranslation(self.Area.height);
				
				Utils.updateDestroyCommonArea(x, z, x1, z1, x2, z2);
			end;
		end;
	else
		g_currentMission:addNodeObject(self.nodeId, self)
	end;
	
	self.FabrikScriptDirtyFlag = self:getNextDirtyFlag();
	
	return true;
end;

function FabrikScript:getSaveAttributesAndNodes(nodeIdent)
	
	local attributes, nodes = "","";
			
	if not self.RunAsGE then
		attributes, nodes = FabrikScript:superClass().getSaveAttributesAndNodes(self, nodeIdent);
	end;
	
	for k,v in pairs (self.Rohstoffe) do
		if 0 < nodes.len(nodes) then
			nodes = nodes .. "\n"
		end
		nodes = nodes..nodeIdent..'<Rohstoff Name="'..v.name..'" Lvl="'..v.fillLevel..'"/>';
	end
	for k,v in pairs (self.Produkte) do
		if 0 < nodes.len(nodes) then
			nodes = nodes .. "\n"
		end
		nodes = nodes..nodeIdent..'<Produkt Name="'..v.name..'" Lvl="'..v.fillLevel..'"/>';
	end
	  
    return attributes,nodes;
end
function FabrikScript:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	
	if not self.RunAsGE and not FabrikScript:superClass().loadFromAttributesAndNodes(self, xmlFile, key, resetVehicles) then
		
		return false
	end
	
	local i = 0
	while true do
		local RohstoffKey = key .. string.format(".Rohstoff(%d)", i)
		if not hasXMLProperty(xmlFile, RohstoffKey) then
			break
		end

		local RohstoffName = getXMLString(xmlFile, RohstoffKey .. "#Name")
		local fillLevel = getXMLFloat(xmlFile, RohstoffKey .. "#Lvl")

		if RohstoffName ~= nil and fillLevel ~= nil and self.Rohstoffe[RohstoffName] ~= nil then
			self:setFillLevel(self.Rohstoffe[RohstoffName], fillLevel)
		end
		i = i + 1
	end	
	i = 0
	while true do
		local ProduktKey = key .. string.format(".Produkt(%d)", i)
		if not hasXMLProperty(xmlFile, ProduktKey) then
			break
		end

		local ProduktName = getXMLString(xmlFile, ProduktKey .. "#Name")
		local fillLevel = getXMLFloat(xmlFile, ProduktKey .. "#Lvl")

		if ProduktName ~= nil and fillLevel ~= nil and self.Produkte[ProduktName] ~= nil then
			self:setFillLevel(self.Produkte[ProduktName], fillLevel)
		end
		i = i + 1
	end	
	return true;
end

function FabrikScript:writeAllStream(streamId, connection)
	for k,v in pairs (self.Rohstoffe) do
		streamWriteInt16(streamId, v.id)
		streamWriteFloat32(streamId, v.fillLevel)
	end
	for k,v in pairs (self.Produkte) do
		streamWriteInt16(streamId, v.id)
		streamWriteFloat32(streamId, v.fillLevel)
	end
	local Fehler = self.Fehler or 0;
	streamWriteInt16(streamId, Fehler)
end;
function FabrikScript:readAllStream(streamId, connection)
	for k,v in pairs (self.Rohstoffe) do
		local id = streamReadInt16(streamId)
		local lvl = streamReadFloat32(streamId)
		self:setFillLevel(self.Rohstoffe[self.StoffeIdToName[id]], lvl)
	end
	for k,v in pairs (self.Produkte) do
		local id = streamReadInt16(streamId)
		local lvl = streamReadFloat32(streamId)
		self:setFillLevel(self.Produkte[self.StoffeIdToName[id]], lvl)
	end
	local Fehler = streamReadInt16(streamId)
	self.Fehler = Fehler > 0 and Fehler or nil;
end; 
function FabrikScript:writeStream(streamId, connection)
	FabrikScript:superClass().writeStream(self, streamId, connection)
	self:writeAllStream(streamId, connection)
end
function FabrikScript:readStream(streamId, connection)
	FabrikScript:superClass().readStream(self, streamId, connection)
	self:readAllStream(streamId, connection)
end;
function FabrikScript:writeUpdateStream(streamId, connection, dirtyMask)
	FabrikScript:superClass().writeUpdateStream(self, streamId, connection, dirtyMask);
	self:writeAllStream(streamId, connection)
end;
function FabrikScript:readUpdateStream(streamId, timestamp, connection)
	FabrikScript:superClass().readUpdateStream(self, streamId, timestamp, connection);
	self:readAllStream(streamId, connection)
end;

function FabrikScript:delete()
	if self.Rohstoffe then
		for k,v in pairs (self.Rohstoffe) do
			if v.TipTrigger and v.TipTrigger.isRegistered then
				v.TipTrigger:unregister()
				v.TipTrigger:delete()
			end
			if v.BaleTrigger then
				removeTrigger(v.BaleTrigger)
			end
			if v.PalletTrigger then
				removeTrigger(v.PalletTrigger)
			end
			if v.WoodTrigger then
				removeTrigger(v.WoodTrigger)
			end
			if v.WaterTrigger then
				v.WaterTrigger:delete()
			end
		end
	end
	if self.Produkte then
		for k,v in pairs (self.Produkte) do
			if v.Heap and v.Heap.ShovelTrigger then
				v.Heap.ShovelTrigger:delete()
			end
			if v.palletSpawner then
				removeTrigger(v.palletSpawner.SpawnerTriggerId)
			end
		end
	end
	if self.PlayerTrigger then
		removeTrigger(self.PlayerTrigger)
	end
	if self.Doors then
		for i=1, table.getn(self.Doors) do
			removeTrigger(self.Doors[i].triggerId);
		end
	end
	for trailer, triggers in pairs(g_currentMission.trailerTipTriggers) do
		if triggers ~= nil then
			for i = 1, table.getn(triggers), 1 do
				if triggers[i] == self then
					table.remove(triggers, i)

					if table.getn(triggers) == 0 then
						g_currentMission.trailerTipTriggers[trailer] = nil
					end
				end
			end
		end
	end
	if self.WorkAnimation then
		if self.WorkAnimation.Ani then
			for index, Ani in pairs(self.WorkAnimation.Ani) do
				if Ani.root then
					delete(Ani.root)
					Ani.root = 0;
				end
			end
		end
	end
	if not self.RunAsGE then FabrikScript:superClass().delete(self) end;
end;

function FabrikScript:update(dt)
	
	if self.isServer then
		for k,v in pairs (self.Produkte) do
			if v.SiloTrigger then
				local trailer = v.SiloTrigger.siloTrailer;
				if v.SiloTrigger.fill >= 4 and trailer ~= nil and (not v.SiloTrigger.fillDone or v.SiloTrigger.notfillDone) then
					trailer:resetFillLevelIfNeeded(v.SiloTrigger.fillType);
					local fillLevel = trailer:getFillLevel(v.SiloTrigger.fillType);
					if v.fillLevel > 0 and trailer:allowFillType(v.SiloTrigger.fillType, false) then
						local deltaFillLevel = math.min(v.SiloTrigger.fillLitersPerSecond*0.001*dt, v.fillLevel);
						local x, y, z = getWorldTranslation(v.SiloTrigger.fillVolumeDischargeInfo.node)
						local d1x, d1y, d1z = localDirectionToWorld(v.SiloTrigger.fillVolumeDischargeInfo.node, v.SiloTrigger.fillVolumeDischargeInfo.width, 0, 0)
						local d2x, d2y, d2z = localDirectionToWorld(v.SiloTrigger.fillVolumeDischargeInfo.node, 0, 0, v.SiloTrigger.fillVolumeDischargeInfo.length)
						local fillSourceStruct = {x = x,y = y,z = z,d1x = d1x,d1y = d1y,d1z = d1z,d2x = d2x,d2y = d2y,d2z = d2z};
						trailer:setFillLevel(fillLevel+deltaFillLevel, v.SiloTrigger.fillType, false, fillSourceStruct);
						local newFillLevel = trailer:getFillLevel(v.SiloTrigger.fillType);
							
						if fillLevel ~= newFillLevel then
							v.fillLevel = math.max(v.fillLevel-(newFillLevel-fillLevel), 0);
							v.SiloTrigger:startFill();
							v.SiloTrigger.CheckSiloTriggerIsFilling = v.SiloTrigger.isFilling;
							self:setIsSiloTriggerFilling(v.SiloTrigger.TriggerId, v.SiloTrigger.isFilling);
						else
							v.SiloTrigger.fillDone = true; -- trailer is full
							v.SiloTrigger:stopFill();
							v.SiloTrigger.CheckSiloTriggerIsFilling = v.SiloTrigger.isFilling;
							self:setIsSiloTriggerFilling(v.SiloTrigger.TriggerId, v.SiloTrigger.isFilling);
						end;
					else
						v.SiloTrigger.fillDone = true -- silo is empty or trailer does not support fill type
						v.SiloTrigger:stopFill();
						v.SiloTrigger.CheckSiloTriggerIsFilling = v.SiloTrigger.isFilling;
						self:setIsSiloTriggerFilling(v.SiloTrigger.TriggerId, v.SiloTrigger.isFilling);
					end;
				elseif v.SiloTrigger.CheckSiloTriggerIsFilling then
					v.SiloTrigger.CheckSiloTriggerIsFilling = nil;
					self:setIsSiloTriggerFilling(v.SiloTrigger.TriggerId, false);
				end;
			end;
		end;
	end;
	
	if self.changeAllowed then
		local RohstoffText = "Rohstoffe:";
		if g_i18n:hasText(self.InputName) then
			RohstoffText = g_i18n:getText(self.InputName)
		elseif g_i18n:hasText("Rohstoffe") then
			RohstoffText = g_i18n:getText("Rohstoffe")
		end
		g_currentMission:addExtraPrintText(RohstoffText);
		for k,v in pairs (self.Rohstoffe) do
			local text = string.format("%s",v.nameL.." [l]");
			local Percentage = math.abs(v.fillLevel / v.capacity * 100);
			text = text.." "..string.format("%d (%d%%)",v.fillLevel ,Percentage);
			g_currentMission:addExtraPrintText(text);
		end
		local ProdukteText = "Produkte:";
		if g_i18n:hasText(self.OutputName) then
			ProdukteText = g_i18n:getText(self.OutputName)
		elseif g_i18n:hasText("Produkte") then
			ProdukteText = g_i18n:getText("Produkte")
		end
		g_currentMission:addExtraPrintText(ProdukteText);
		for k,v in pairs (self.Produkte) do
			local text = string.format("%s",v.nameL.." [l]");
			local Percentage = math.abs(v.fillLevel / v.capacity * 100);
			text = text.." "..string.format("%d (%d%%)",v.fillLevel ,Percentage);
			g_currentMission:addExtraPrintText(text);
		end
		if self.Fehler then
			g_currentMission:addExtraPrintText((g_i18n:hasText("ERROR") and g_i18n:getText("ERROR") or "ERROR: ")..self.FehlerText[self.Fehler]);
		end
		if not g_currentMission.controlPlayer then
			self.changeAllowed = false;
		end;
	end;
end;

function FabrikScript:updateTick(dt)
	
	self.updateMs = self.updateMs + (dt * g_currentMission.missionStats.timeScale);
	if self.updateMs >= 60000   then
		self.updateMs = self.updateMs - 60000;
		self.updateMin = self.updateMin + 1;
		if self.updateMin >= self.updateIntervall then
			self.updateMin = self.updateMin - self.updateIntervall
			if self.isServer then
				local need = (self.ProduktPerHour/60 * self.updateIntervall) --Bedarf pro intervall 
				local RohstoffeVorhanden = true;
				for k,v in pairs (self.Rohstoffe) do
					if v.fillLevel < need * v.factor then
						need = v.fillLevel / v.factor
						if need <= 0 then
							RohstoffeVorhanden = false;
							self.Fehler = v.id
							break;
						end;
					end;
				end
				local CapacityVorhanden = true;
				for k,v in pairs (self.Produkte) do
					if v.palletSpawner ~= nil then
						if v.palletSpawner.currentPallet then
							if not entityExists(v.palletSpawner.currentPallet.nodeId) or getName(getParent(v.palletSpawner.currentPallet.nodeId)) ~= "RootNode" then
								v.palletSpawner.numObjectsInPalletSpawnerTrigger = v.palletSpawner.numObjectsInPalletSpawnerTrigger - (v.palletSpawner.numObjectsPerID[v.palletSpawner.currentPallet.nodeId] or 0);
								v.palletSpawner.numObjectsPerID[v.palletSpawner.currentPallet.nodeId] = nil;
								v.palletSpawner.currentPallet = nil;	
							end;
						end;
						if v.palletSpawner.currentPallet == nil then
							v.fillLevel = 0;
							if v.palletSpawner.numObjectsInPalletSpawnerTrigger > 0 then -- v1.3.4 fix Fremd Objekt in PalletSpawnerTrigger (inspired TracMax)
								CapacityVorhanden = false;
								self.Fehler = v.id + 1;
								break;
							end;
						end;
					end;
					if v.fillLevel + need * v.factor > v.capacity then
						need = (v.capacity-v.fillLevel) / v.factor
						if need <= 0 then
							CapacityVorhanden = false;
							self.Fehler = v.id
							break;
						end
					end;
				end
				if RohstoffeVorhanden and CapacityVorhanden then
					for k,v in pairs (self.Rohstoffe) do
						self:setFillLevel(v, v.fillLevel - need * v.factor)
					end
					for k,v in pairs (self.Produkte) do
						if v.palletSpawner ~= nil then
							if v.palletSpawner.currentPallet == nil and v.palletSpawner.numObjectsInPalletSpawnerTrigger == 0 then
								local x, y, z = getWorldTranslation(v.palletSpawner.SpawnerPlaceId)
								local rx, ry, rz = getWorldRotation(v.palletSpawner.SpawnerPlaceId)
								local pallet = FillablePallet:new(self.isServer, self.isClient)

								if pallet.load(pallet, v.palletSpawner.Filename, x, y, z, rx, ry, rz) then
									pallet.register(pallet)

									v.palletSpawner.currentPallet = pallet
									v.palletSpawner.currentPallet.isReferenced = true
									v.capacity = pallet:getCapacity();
								else
									pallet.delete(pallet)
								end
							end

							if v.palletSpawner.currentPallet ~= nil then
								v.palletSpawner.currentPallet:setFillLevel(v.palletSpawner.currentPallet:getFillLevel() + need * v.factor);
								v.fillLevel = v.palletSpawner.currentPallet:getFillLevel();
							end
						else
							self:setFillLevel(v, v.fillLevel + need * v.factor)
						end
					end
					self.Fehler = nil;
				else
					self.SendUpdate = true;
				end
			end;
			if self.isClient and self.WorkAnimation ~= nil then
				if self.Fehler == nil then
					if self.WorkAnimation.PS then
						local numPS = table.getn(self.WorkAnimation.PS)
						for i = 1, numPS do
							local PS = self.WorkAnimation.PS[i];
							if PS.isEmitting ~= true and PS.elapsed == nil then
								Utils.setEmittingState(PS, true)
							end
						end
					end;
					if self.WorkAnimation.Ani then
						local numAni = table.getn(self.WorkAnimation.Ani)
						for i = 1, numAni do
							local Ani = self.WorkAnimation.Ani[i];
							if not isAnimTrackEnabled(Ani.Animi, 0) then
								enableAnimTrack(Ani.Animi, 0);
							end;
						end;
					end;
					if self.WorkAnimation.Sound then
						local numSound = table.getn(self.WorkAnimation.Sound)
						for i = 1, numSound do
							local Sound = self.WorkAnimation.Sound[i];
							if not getVisibility(Sound.node) and Sound.elapsed == nil then
								setVisibility(Sound.node,true)
							end;
						end;
					end;
					if self.WorkAnimation.Shader then
						local numShader = table.getn(self.WorkAnimation.Shader)
						for i = 1, numShader do
							local SP = self.WorkAnimation.Shader[i];
							if not SP.shared then
								SP.shared = true;
								setShaderParameter(SP.node,SP.Parameter,SP.on.x,SP.on.y,SP.on.z,SP.on.w,false)	
							end;
						end;
					end;
					self.StopWorkAni = nil;
				else
					self.StopWorkAni = true;
				end
			end
		end;
	end;
	
	if self.isClient then
		if self.WorkAnimation ~= nil and self.Fehler == nil then
			if self.WorkAnimation.Ani then
				local numAni = table.getn(self.WorkAnimation.Ani)
				for i = 1, numAni do
					local Ani = self.WorkAnimation.Ani[i];
					if isAnimTrackEnabled(Ani.Animi, 0) then
						if getAnimTrackTime(Ani.Animi, 0) >= getAnimClipDuration(Ani.Animi, 0) then
							setAnimTrackTime(Ani.Animi, 0, 0, false);
						end;
					end;
				end;
			end;
			if self.WorkAnimation.PS then
				local numPS = table.getn(self.WorkAnimation.PS)
				for i = 1, numPS do
					local PS = self.WorkAnimation.PS[i];
					if PS.elapsed then
						PS.elapsed = PS.elapsed - dt;
						if PS.elapsed <= 0 then
							PS.elapsed = PS.elapsed + PS.intervall;
							Utils.setEmittingState(PS, not PS.isEmitting)
						end
					end
				end
			end;
			if self.WorkAnimation.Sound then
				local numPS = table.getn(self.WorkAnimation.Sound)
				for i = 1, numPS do
					local Sound = self.WorkAnimation.Sound[i];
					if Sound.elapsed then
						Sound.elapsed = Sound.elapsed - dt;
						if Sound.elapsed <= 0 then
							Sound.elapsed = Sound.elapsed + Sound.intervall;
							setVisibility(Sound.node,not getVisibility(Sound.node))
						end;
					end;
				end;
			end;
		elseif self.WorkAnimation ~= nil and self.StopWorkAni then
			local AniDisabled = true;
			if self.WorkAnimation.Ani then
				local numAni = table.getn(self.WorkAnimation.Ani)
				for i = 1, numAni do
					local Ani = self.WorkAnimation.Ani[i];
					if isAnimTrackEnabled(Ani.Animi, 0) then
						if Ani.IgnoreDuration then
							disableAnimTrack(Ani.Animi, 0);
							setAnimTrackTime(Ani.Animi, 0, 0, false);
						elseif getAnimTrackTime(Ani.Animi, 0) >= getAnimClipDuration(Ani.Animi, 0) then
							disableAnimTrack(Ani.Animi, 0);
							setAnimTrackTime(Ani.Animi, 0, 0, false);
						else
							AniDisabled = false;
						end;
					end;
				end;
			end;
			if AniDisabled then
				if self.WorkAnimation.PS then
					local numPS = table.getn(self.WorkAnimation.PS)
					for i = 1, numPS do
						local PS = self.WorkAnimation.PS[i];
						if PS.isEmitting == true then
							Utils.setEmittingState(PS, false)
						end
					end
				end;
				if self.WorkAnimation.Sound then
					local numSound = table.getn(self.WorkAnimation.Sound)
					for i = 1, numSound do
						local Sound = self.WorkAnimation.Sound[i];
						if getVisibility(Sound.node) then
							setVisibility(Sound.node,false)
						end;
					end;
				end;
				if self.WorkAnimation.Shader then
					local numShader = table.getn(self.WorkAnimation.Shader)
					for i = 1, numShader do
						local SP = self.WorkAnimation.Shader[i];
						if SP.shared then
							SP.shared = false;
							setShaderParameter(SP.node,SP.Parameter,SP.off.x,SP.off.y,SP.off.z,SP.off.w,false)	
						end;
					end;
				end;
				self.StopWorkAni = nil;
			end;
		end;
	end;
	
	if self.Doors then
		for i=1, table.getn(self.Doors) do
			local old = self.Doors[i].Trans;
			if (self.Doors[i].entred > 0) then
				if self.Doors[i].Trans < self.Doors[i].maxTrans then
					self.Doors[i].Trans = math.min(self.Doors[i].Trans + dt*self.Doors[i].transTime, self.Doors[i].maxTrans);
				end;
			elseif (self.Doors[i].entred <= 0) then
				if self.Doors[i].Trans > self.Doors[i].minTrans then
					self.Doors[i].Trans = math.max(self.Doors[i].Trans - dt*self.Doors[i].transTime, self.Doors[i].minTrans);
				end;
			end;

			if old ~= self.Doors[i].Trans then
				local dir = 1;
				for j=1, table.getn(self.Doors[i].door) do
					local x, y, z = getTranslation(self.Doors[i].door[j]);
					setTranslation(self.Doors[i].door[j], x, y, self.Doors[i].Trans * dir);
					dir = dir * -1
				end;
			end;
		end;
	end;
	
	if self.isServer and self.SendUpdate then
		self.SendUpdate = nil
		self:raiseDirtyFlags(self.FabrikScriptDirtyFlag);
	end
end;

function FabrikScript:allowFillType(art,FillType)
	return art.acceptedFillTypes[FillType];
end

function FabrikScript:getTipInfoForTrailer(trailer, tipReferencePointIndex)
    Debug(8,string.format("getTipInfoForTrailer art %s trailer %s  tipReferencePointIndex %s",tostring(art),tostring(trailer), tostring(tipReferencePointIndex)));
	local isAllowed, minDistance, bestPoint = true, math.huge, nil;
	if self.trailerTipTrigger[trailer] and self.trailerTipTrigger[trailer].TipTrigger then
		isAllowed, minDistance, bestPoint = self.trailerTipTrigger[trailer].TipTrigger:getTipInfoForTrailer(trailer, tipReferencePointIndex);
	end
    return isAllowed, minDistance, bestPoint;
end

function FabrikScript:updateTrailerTipping(trailer, fillDelta, fillType)
	Debug(8,string.format("updateTrailerTipping art %s trailer %s fillDelta %s fillType %s",tostring(art),tostring(trailer), tostring(fillDelta), tostring(fillType)));
	if self.trailerTipTrigger[trailer] and self.trailerTipTrigger[trailer].TipTrigger then
		self.trailerTipTrigger[trailer].TipTrigger:updateTrailerTipping(trailer, fillDelta, fillType)
	end;
end;

function FabrikScript:setIsWaterTankFilling(WasserTriggerId, isFilling, trailer,  noEventSend)
	if self.Rohstoffe then
		for k,v in pairs (self.Rohstoffe) do
			if v.WaterTrigger and v.WaterTrigger.WasserTriggerId and v.WaterTrigger.WasserTriggerId == WasserTriggerId then
				v.WaterTrigger:setIsWaterTankFilling(isFilling, trailer, noEventSend)
				break;
			end
		end
	end
end

function FabrikScript:setIsSiloTriggerFilling(TriggerId, isFilling, noEventSend)
	if self.isServer then
		FS_SiloTriggerSetIsFillingEvent.sendEvent(self, TriggerId, isFilling, noEventSend)
	end
	if self.isClient then	
		for k,v in pairs (self.Produkte) do
			if v.SiloTrigger then
				if v.SiloTrigger.isFilling ~= isFilling then
					if isFilling then
						v.SiloTrigger:startFill();
					else
						v.SiloTrigger:stopFill();
					end;
				end;
			end;
		end;
	end;
end

function FabrikScript:getCapacity(art)
	return art.capacity;
end

function FabrikScript:setFillLevel(art, fillLevel, fillType)
	Debug(10,string.format("setFillLevel fillLevel %.2f  fillType %s",fillLevel, tostring(fillType)));
	local oldLvl = art.fillLevel
	art.fillLevel = fillLevel
	if art.TipTrigger then 
		art.TipTrigger:updateMoving(fillLevel-oldLvl)
	elseif art.Heap then
		self:onAmountChanged(art.Heap,fillLevel)
	elseif art.WaterTrigger then
		self:onAmountChanged(art.WaterTrigger,fillLevel)
	elseif art.logs then
		self:updateWoodTriggerLogs(art,fillLevel)
	end;
	self.SendUpdate = true;
end

function FabrikScript:getFillLevel(art,FillType)
	return art.fillLevel;
end

function FabrikScript:resetFillLevelIfNeeded(art)
end

function FabrikScript:addShovelFillLevel(target, shovel, fillLevel, fillType)
	Debug(10,string.format("addShovelFillLevel fillLevel %.2f  fillType %d",fillLevel, fillType));
	if target.Rohstoff and target.fillTypes[fillType] then
		local rLvl = self.Rohstoffe[target.Rohstoff].fillLevel;
		local rCap = self.Rohstoffe[target.Rohstoff].capacity
		Debug(10,string.format("addShovelFillLevel rLvl %.2f < rCap %.2f",rLvl, rCap));
		if rLvl < rCap then
	        fillLevel = math.min(fillLevel, rCap - rLvl);
			self:setFillLevel(self.Rohstoffe[target.Rohstoff],rLvl + fillLevel, fillType)
			return fillLevel;
		end;
	end;
    return 0;
end;

function FabrikScript:fillShovel(ST, shovel, dt)
	if ST.Produkt then
		local fillLevel = self:getFillLevel(self.Produkte[ST.Produkt],ST.fillType);
		if fillLevel > 0 then
			local delta = shovel:fillShovelFromTrigger(ST, fillLevel, ST.fillType, dt);
			if delta > 0 then
				self:setFillLevel(self.Produkte[ST.Produkt],fillLevel-delta, ST.fillType);
			end;
		end;
	end;
end;

function FabrikScript:onAmountChanged(moving,amount)
    if moving.moveMaxAmount then
		if amount < 0.001 then
			amount = 0;
		end;
		local mover = 0;
		if moving.movingIndex then
			mover = moving.movingIndex
		elseif moving.id then
			mover = moving.id
		end
		local x,y,z = getTranslation(mover);
		local y = moving.moveMinY + (moving.moveMaxY - moving.moveMinY)*Utils.clamp(amount, 0, moving.moveMaxAmount)/(moving.moveMaxAmount);
		setTranslation(mover, x,y,z);
	end;
end;

function FabrikScript:updateWoodTriggerLogs(Trigger,fillLevel)
	if self.isClient and Trigger.logs then
		local numVisibilityNodes = table.getn(Trigger.logs)
		local numVisible = math.ceil((numVisibilityNodes*fillLevel)/Trigger.capacity)
		for i = 1, numVisibilityNodes, 1 do
			setVisibility(Trigger.logs[i].node, i <= numVisible)
			setRigidBodyType(Trigger.logs[i].node, i <= numVisible and Trigger.logs[i].rigidBody or "NoRigidBody")
		end
	end
end

function FabrikScript:PlayerTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
	
	if (g_currentMission.controlPlayer and g_currentMission.player and otherId == g_currentMission.player.rootNode) then
		if (onEnter) then 
            self.changeAllowed = true;
        elseif (onLeave) then
            self.changeAllowed = false;
        end;
	end;
	
end;

function FabrikScript:doorTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	
	for i=1, table.getn(self.Doors) do
		if self.Doors[i].triggerId == triggerId then
			if onEnter then
				self.Doors[i].entred = self.Doors[i].entred + 1
			else -- onLeave
				self.Doors[i].entred = math.max(self.Doors[i].entred - 1,0)
			end;
			break;
		end;
	end;

end;

function FabrikScript:BaleTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter and otherActorId ~= 0 then
		local object = g_currentMission:getNodeObject(otherActorId);
		if object ~= nil and object:isa(Bale) then
			local fillLevel = object:getFillLevel();
			local fillType = object:getFillType();
			for k,v in pairs (self.Rohstoffe) do
				if v.BaleTrigger == triggerId and self:allowFillType(v,fillType) then
					if v.fillLevel + fillLevel <= v.capacity then
						self:setFillLevel(v,v.fillLevel + fillLevel,fillType)
						object:delete();  
					end;   
					break; 
				end;
			end;
		end;
	end;
end;

function FabrikScript:PalletTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter and otherActorId ~= 0 then
		local object = g_currentMission:getNodeObject(otherActorId)
		if object ~= nil and object.isa(object, FillablePallet) then
			if g_currentMission:getIsServer() then
				local fillType = object:getFillType()
				for k,v in pairs (self.Rohstoffe) do
					if v.PalletTrigger == triggerId and self:allowFillType(v,fillType) then
						local fillLevel = object:getFillLevel()
						if v.fillLevel + fillLevel <= v.capacity then
							self:setFillLevel(v,v.fillLevel + fillLevel,fillType)
							object:delete()
						end;
						break;
					end;
				end;
			end
		end
	end
end

function FabrikScript:WoodTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter and otherActorId ~= 0 then
		local splitType = SplitUtil.splitTypes[getSplitType(otherActorId)]

		if splitType ~= nil and 0 < splitType.woodChipsPerLiter then
			
			for k,v in pairs (self.Rohstoffe) do
				if v.WoodTrigger == triggerId then --and self:allowFillType(v,fillType) then
					if g_currentMission:getIsServer() then
						local volume = getVolume(otherActorId)
						
						local fillLevel = volume*1000*splitType.woodChipsPerLiter
						Debug(5,string.format("wood %s Lvl %.2f volume %.2f woodChipsPerLiter %.2f ",splitType.name,fillLevel,volume,splitType.woodChipsPerLiter));
						if v.fillLevel + fillLevel <= v.capacity then
							self:setFillLevel(v,v.fillLevel + fillLevel,fillType)
							delete(otherActorId)
						end
					end
					break;
				end
			end
		end
	end;
end;

function FabrikScript:palletSpawnerTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	for k,v in pairs (self.Produkte) do
		if v.palletSpawner and v.palletSpawner.SpawnerTriggerId == triggerId then --and self:allowFillType(v,fillType) then
			if onEnter then
				v.palletSpawner.numObjectsInPalletSpawnerTrigger = v.palletSpawner.numObjectsInPalletSpawnerTrigger + 1
				v.palletSpawner.numObjectsPerID[otherActorId] = (v.palletSpawner.numObjectsPerID[otherActorId] or 0) + 1;
				if v.palletSpawner.currentPallet == nil and otherActorId ~= 0 then
					local object = g_currentMission:getNodeObject(otherActorId)

					if object ~= nil and object.isa(object, FillablePallet) and self:allowFillType(v,object:getFillType()) then
						v.palletSpawner.currentPallet = object
						v.palletSpawner.currentPallet.isReferenced = true
						v.capacity = object:getCapacity();
						v.fillLevel = object:getFillLevel();
					end
				end
			elseif onLeave then
				v.palletSpawner.numObjectsInPalletSpawnerTrigger = v.palletSpawner.numObjectsInPalletSpawnerTrigger - 1
				v.palletSpawner.numObjectsPerID[otherActorId] = (v.palletSpawner.numObjectsPerID[otherActorId] or 0) - 1;
				if v.palletSpawner.numObjectsPerID[otherActorId] <= 0 then
					v.palletSpawner.numObjectsPerID[otherActorId] = nil;
				end
				if v.palletSpawner.currentPallet ~= nil and v.palletSpawner.currentPallet.nodeId == otherActorId then
					v.palletSpawner.currentPallet.isReferenced = false
					v.palletSpawner.currentPallet = nil
					v.fillLevel = 0;
				end
			end
		end;
	end;
end

function FabrikScript:TipTriggerCallback(art, triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	Debug(8,string.format("TipTriggerCallback art %s riggerId %s otherId %s onEnter %s onLeave %s onStay %s otherShapeId %s",tostring(art), tostring(triggerId), tostring(otherId), tostring(onEnter), tostring(onLeave), tostring(onStay), tostring(otherShapeId)));
	
	local trailer = g_currentMission.objectToTrailer[otherShapeId];
	if trailer ~= nil and trailer.allowTipDischarge then
		if onEnter then
			if g_currentMission.trailerTipTriggers[trailer] == nil then
				g_currentMission.trailerTipTriggers[trailer] = {};
			end;
			self.trailerTipTrigger[trailer] = art;
			table.insert(g_currentMission.trailerTipTriggers[trailer], self);
		elseif onLeave then
			local triggers = g_currentMission.trailerTipTriggers[trailer];
			if triggers ~= nil then
				for i=1, table.getn(triggers) do
					if triggers[i] == self then
						self.trailerTipTrigger[trailer] = nil;
						table.remove(triggers, i);
						if table.getn(triggers) == 0 then
							g_currentMission.trailerTipTriggers[trailer] = nil;
						end;
						break;
					end;
				end;
			end;
		end;
	end;
end;
 
if FabrikScript.RunAsGE then
	g_onCreateUtil.addOnCreateFunction("FabrikScript", FabrikScript.onCreate);
else
	registerPlaceableType("FabrikScript", FabrikScript);
end;
 
FS_SiloTriggerSetIsFillingEvent = {}
FS_SiloTriggerSetIsFillingEvent_mt = Class(FS_SiloTriggerSetIsFillingEvent, Event)
InitEventClass(FS_SiloTriggerSetIsFillingEvent, "FS_SiloTriggerSetIsFillingEvent")
function FS_SiloTriggerSetIsFillingEvent:emptyNew()
	local self = Event:new(FS_SiloTriggerSetIsFillingEvent_mt)
	return self
end
function FS_SiloTriggerSetIsFillingEvent:new(object, TriggerId, isFilling)
	local self = FS_SiloTriggerSetIsFillingEvent:emptyNew()
	self.object = object
	self.TriggerId = TriggerId;
	self.isFilling = isFilling
	return self
end
function FS_SiloTriggerSetIsFillingEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId))
	self.TriggerId = streamReadInt32(streamId);
	self.isFilling = streamReadBool(streamId)
	self:run(connection)
end
function FS_SiloTriggerSetIsFillingEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object))
	streamWriteInt32(streamId, self.TriggerId)
	streamWriteBool(streamId, self.isFilling)
end
function FS_SiloTriggerSetIsFillingEvent:run(connection)
	if self.object ~= nil then
		self.object:setIsSiloTriggerFilling(self.TriggerId, self.isFilling, true)
	end;
end
function FS_SiloTriggerSetIsFillingEvent.sendEvent(object, TriggerId, isFilling, noEventSend)
	if (noEventSend == nil or noEventSend == false) then
		if g_server ~= nil then
			g_server:broadcastEvent(FS_SiloTriggerSetIsFillingEvent:new(object, TriggerId, isFilling), nil, nil, object)
		else
			g_client:getServerConnection():sendEvent(FS_SiloTriggerSetIsFillingEvent:new(object, TriggerId, isFilling))
		end
	end
end
